
/* -*-mode:C;c-basic-offset:8-*-
 *      Copyright (C) 1993-1995 Bas Laarhoven.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 $Source: /homes/cvs/zftape/ftape-ctl.c,v $
  $Author: claus $
  *
 $Revision: 1.25 $
 $Date: 1996/07/22 13:27:07 $
 $State: Exp $
 *
 *      This file contains the non-read/write ftape functions
 *      for the QIC-40/80 floppy-tape driver for Linux.
 */

static char RCSid[] =
"$Id: ftape-ctl.c,v 1.25 1996/07/22 13:27:07 claus Exp $";

  
#include "ftape.h"
#include <linux/errno.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <linux/string.h>

#include "ftape-eof.h"
#include "ftape-io.h"
#include "ftape-ctl.h"
#include "ftape-write.h"
#include "ftape-read.h"
#include "ftape-rw.h"
#include "qic117.h"
#include "ftape-bsm.h"
#include "qic80-vtbl.h"
#include "qic80-compress.h"
#include "zftape-ioctl.h"


/*      Global vars.
 */
int segments_per_track = 102;
int segments_per_head = 1020;
int segments_per_cylinder = 4;
int tracks_per_tape = 20;
int ftape_failure = 1;
int ftape_seg_pos = 0;
int first_data_segment = -1;
int ftape_last_data_segment = -1;
int ftape_state = idle;         /* use buffer_state_enum */
history_record history;
int write_protected;
vendor_struct drive_type;
int ftape_offline = 0;
int no_tape = 1;
int formatted = 0;
int ftape_data_rate = 0;
int going_offline = 0;
int read_only = 0;
int ftape_block_size = DEFAULT_BLOCK_SIZE;
int ftape_resid = 0;


/*      Local vars.
 */
static int ftape_last_error = 0;
static const vendor_struct vendors[] = QIC117_VENDORS;
static const wakeup_method methods[] = WAKEUP_METHODS;
static int init_drive_needed = 1;


typedef int (mt_fun)( int *argptr );
typedef int (*mt_funp)( int *argptr );
typedef struct
{
	mt_funp function;
	unsigned offline         : 1; /* op permitted if offline or no_tape */
	unsigned write_protected : 1; /* op permitted if write-protected    */
	unsigned not_formatted   : 1; /* op permitted if tape not formatted */
	unsigned raw_mode        : 1; /* op permitted if qic80_mode == 0    */
	unsigned need_idle_state : 1; /* need to call def_idle_state        */
} fun_entry;

static mt_fun mt_dummy, mt_reset, mt_fsr, mt_bsr, mt_rew, mt_offl, mt_nop,
	mt_weof, mt_erase, mt_ras2, mt_setblk, mt_setdensity,
	mt_seek, mt_tell, mt_reten, mt_eom, mt_fsf, mt_bsf,
	mt_fsfm, mt_bsfm, mt_setdrvbuffer;

static fun_entry mt_funs[]=
{ 
	{mt_reset       , 1, 1, 1, 1, 0 }, /* MT_RESET  0 */
	{mt_fsf         , 0, 1, 0, 0, 1 }, /* MT_FSF      */
	{mt_bsf         , 0, 1, 0, 0, 1 }, /* MT_BSF      */
	{mt_fsr         , 0, 1, 0, 1, 1 }, /* MT_FSR      */
	{mt_bsr         , 0, 1, 0, 1, 1 }, /* MT_BSR      */
	{mt_weof        , 0, 0, 0, 0, 1 }, /* MT_WEOF   5 */
	{mt_rew         , 0, 1, 1, 1, 0 }, /* MT_REW      */
	{mt_offl        , 0, 1, 1, 1, 0 }, /* MT_OFFL     */
	{mt_nop         , 1, 1, 1, 1, 1 }, /* MT_NOP      */
	{mt_reten       , 0, 1, 1, 1, 0 }, /* MT_RETEN    */
	{mt_bsfm        , 0, 1, 0, 0, 1 }, /* MT_BSFM  10 */
	{mt_fsfm        , 0, 1, 0, 0, 1 }, /* MT_FSFM     */
	{mt_eom         , 0, 1, 0, 0, 1 }, /* MT_EOM      */
	{mt_erase       , 0, 0, 0, 1, 0 }, /* MT_ERASE    */
	{mt_dummy       , 1, 1, 1, 1, 0 }, /* MT_RAS1     */
	{mt_ras2        , 0, 0, 0, 1, 0 }, /* MT_RAS2 ftape_fix */
	{mt_dummy       , 1, 1, 1, 1, 0 }, /* MT_RAS3     */
	{mt_dummy       , 1, 1, 1, 1, 0 }, /*      ?      */
	{mt_dummy       , 1, 1, 1, 1, 0 }, /*      ?      */
	{mt_dummy       , 1, 1, 1, 1, 0 }, /*      ?      */
	{mt_setblk      , 1, 1, 1, 1, 1 }, /* MT_SETBLK 20*/
	{mt_setdensity  , 1, 1, 1, 1, 0 }, /* MT_SETDENSITY  set trace-level */
	{mt_seek        , 0, 1, 0, 1, 1 }, /* MT_SEEK     */
	{mt_dummy       , 0, 1, 0, 1, 1 }, /* MT_TELL     */ /* MTIOCTOP are write-only */
	{mt_setdrvbuffer, 1, 1, 1, 1, 0 }, /* MT_SETDRVBUFFER:
					    * change number of dma-buffers 
					    */
	{mt_dummy       , 1, 1, 1, 1, 0 }, /* MT_FSS   25 */
	{mt_dummy       , 1, 1, 1, 1, 0 }, /* MT_BSS      */
	{mt_dummy       , 1, 1, 1, 1, 0 }  /* MT_WSM      */
};  

#define NR_MT_CMDS NR_ITEMS(mt_funs)

static int
ftape_not_operational ( int status )
{
	/* return true if status indicates tape can not be used.
	 */
	return ( (status ^ QIC_STATUS_CARTRIDGE_PRESENT) &
		 (QIC_STATUS_ERROR |
		  QIC_STATUS_CARTRIDGE_PRESENT |
		  QIC_STATUS_NEW_CARTRIDGE) );
}

int
ftape_seek_to_eot( void)
{
	TRACE_FUN( 5, "ftape_seek_to_eot");
	int result;
	int status;

	result = ftape_ready_wait( timeout.pause, &status);
	while ((status & QIC_STATUS_AT_EOT) == 0) {
		if (result < 0) {
			TRACE( 1, "failed");
			TRACE_EXIT;
			return result;
		}
		if (ftape_not_operational( status)) {
			TRACE_EXIT;
			return -EIO;
		}
		result = ftape_command_wait( QIC_PHYSICAL_FORWARD,
					     timeout.rewind, &status);
	}
	TRACE_EXIT;
	return 0;
}

int
ftape_seek_to_bot( void)
{
	TRACE_FUN( 5, "ftape_seek_to_bot");
	int result;
	int status;

	result = ftape_ready_wait( timeout.pause, &status);
	while ((status & QIC_STATUS_AT_BOT) == 0) {
		if (result < 0) {
			TRACE( 1, "failed");
			TRACE_EXIT;
			return result;
		}
		if (ftape_not_operational( status)) {
			TRACE_EXIT;
			return -EIO;
		}
		result = ftape_command_wait( QIC_PHYSICAL_REVERSE,
					     timeout.rewind, &status);
	}
	TRACE_EXIT;
	return 0;
}

void
ftape_reset_position( void)
{
	if ( zftape_qic80_mode && !ftape_old_ftape ) {
		/*
		 * need to keep track of the volume table and compression map
		 */
		if ( zftape_compression_map_location != 0 ) {
			ftape_seg_pos = ftape_first_user_segment;
		} else {
			ftape_seg_pos = first_data_segment + 1;
		}
	} else {
		ftape_seg_pos = first_data_segment;
	}
	ftape_seg_data_pos =
		ftape_uncmpr_pos   = 0;
	if ( first_data_segment != -1 ) {
		/*
		 * need to keep track of the volume table and compression map
		 */
		ftape_data_pos = ftape_calc_data_pos( ftape_seg_pos, 
						      ftape_seg_data_pos );
	} else {
		ftape_data_pos     = 0;
	}
	just_before_eof = 0;
	zftape_reset_cmpr_locals();
}

int
ftape_new_cartridge( void)
{
	location.track = -1;          /* force seek on first access */
	first_data_segment = -1;      /* unknown */
	ftape_first_user_segment = -1;
	ftape_resid = 0;
	ftape_io_state = io_idle;
	ftape_block_size = DEFAULT_BLOCK_SIZE;
	ftape_zap_read_buffers();
	ftape_zap_write_buffers();
	ftape_reset_position();
	qic80_init_vtbl();
	return 0;
}

int
ftape_abort_operation( void)
{
	TRACE_FUN( 5, "ftape_abort_operation");
	int result = 0;
	int i;
	int status;

	if (runner_status == running) {
		TRACE( 5, "aborting runner, waiting");
		runner_status = do_abort;
		/* set timeout so that the tape will run to logical EOT
		 * if we missed the last sector and there are no queue pulses.
		 */
		result = ftape_dumb_stop();
		if (result == 0) {
			runner_status = idle;
		}
	}
	if (runner_status != idle) {
		if (runner_status == do_abort) {
			TRACE( 5, "forcing runner abort");
		}
		TRACE( 5, "stopping tape");
		result = ftape_command_wait( QIC_STOP_TAPE, timeout.stop, &status);
		location.known = 0;
		runner_status = idle;
	}
	for (i = 0; i < ftape_num_buffers; ++i) {
		buffer[ i].status = waiting;
	}
	head = tail = 0;
	TRACE_EXIT;
	return result;
}

int
lookup_vendor_id( int vendor_id)
{
	int i = 0;

	while ( vendors[ i].vendor_id != vendor_id) {
		if (++i >= NR_ITEMS( vendors)) {
			return -1;
		}
	}
	return i;
}

void
ftape_detach_drive( void)
{
	TRACE_FUN( 8, "ftape_detach_drive");

	TRACE( 5, "disabling tape drive and fdc");
	ftape_put_drive_to_sleep( drive_type);
	fdc_catch_stray_interrupts( 1);   /* one always comes */
	fdc_disable();
	fdc_release_irq_and_dma();
	TRACE_EXIT;
}

static void
clear_history( void)
{
	history.used = 0;
	history.id_am_errors =
		history.id_crc_errors =
		history.data_am_errors =
		history.data_crc_errors =
		history.overrun_errors =
		history.no_data_errors =
		history.retries =
		history.crc_errors =
		history.crc_failures =
		history.ecc_failures =
		history.corrected =
		history.defects =
		history.rewinds =
		history.cmpr_time =
		history.wr_uncompressed =
		history.wr_compressed =
		history.rd_uncompressed =
		history.rd_compressed = 0;
}

int
ftape_activate_drive( vendor_struct* drive_type)
{
	TRACE_FUN( 5, "ftape_activate_drive");
	int result = 0;

	/* If we already know the drive type, wake it up.
	 * Else try to find out what kind of drive is attached.
	 */
	if (drive_type->wake_up != unknown_wake_up) {
		TRACE( 5, "enabling tape drive and fdc");
		result = ftape_wakeup_drive( drive_type->wake_up);
		if (result < 0) {
			TRACE( 1, "known wakeup method failed");
		}
	} else {
#ifndef NO_TRACE_AT_ALL
		int old_tracing = tracing;
#endif
		wake_up_types method;

		/*  Try to awaken the drive using all known methods.
		 *  Lower tracing for a while.
		 */       
#ifndef NO_TRACE_AT_ALL
		if (tracing <= 4) {
			tracing = 0;
		}                  
#endif
		for (method = no_wake_up; method < NR_ITEMS( methods); ++method) {
			drive_type->wake_up = method;
			/*  Test setup for dual drive configuration in dodo.
			 *  /dev/rft2 uses mountain wakeup only -> Archive QIC-80
			 *  /dev/rft3 uses colorado wakeup only -> Jumbo QIC-40
			 *  Other systems will use the normal scheme.
			 */
#ifdef ANAXAGORAS
			if (    (FTAPE_UNIT < 2) 
				|| (FTAPE_UNIT == 2 && method == wake_up_mountain) 
				|| (FTAPE_UNIT == 3 && method == wake_up_colorado)) {
				result = ftape_wakeup_drive( drive_type->wake_up);
			} else {
				result = -EIO;
			}
#else
			result = ftape_wakeup_drive( drive_type->wake_up);
#endif
			if (result >= 0) {
#ifndef NO_TRACE_AT_ALL
				int tracing = old_tracing; /* fool TRACE */
#endif
				TRACEx1( 4, "drive wakeup method: %s",
					 methods[ drive_type->wake_up].name);
				break;
			}
		} 
#ifndef NO_TRACE_AT_ALL
		tracing = old_tracing;
#endif
		if (method >= NR_ITEMS( methods)) {
			/* no response at all, cannot open this drive */
			drive_type->wake_up = unknown_wake_up;
			TRACE( 1, "no tape drive found !");
#ifndef NO_TRACE_AT_ALL
			tracing = old_tracing;
#endif
			result = -ENODEV;
		}
	}
	TRACE_EXIT;
	return result;
}

int
ftape_get_drive_status( int* new_tape, int* no_tape, int* wp_tape)
{
	TRACE_FUN( 5, "ftape_get_drive_status");
	int result;
	int status;

	*no_tape =
		*wp_tape = 0;
	/*    Tape drive is activated now.
	 *    First clear error status if present.
	 */
	do {
		result = ftape_ready_wait( timeout.reset, &status);
		if (result < 0) {
			if (result == -ETIME) {
				TRACE( 1, "ftape_ready_wait timeout");
			} else if (result == -EINTR) {
				TRACE( 1, "ftape_ready_wait aborted");
			} else {
				TRACE( 1, "ftape_ready_wait failed");
			}
			result = -EIO;
			break;
		}
		/*  Clear error condition (drive is ready !)
		 */
		if (status & QIC_STATUS_ERROR) {
			int error;
			int command;

			TRACE( 1, "error status set");
			result = ftape_report_error( &error, &command, 1);
			if (result < 0) {
				TRACEi( 1, "report_error_code failed:", result);
				ftape_reset_drive();    /* hope it's working next time */
				init_drive_needed = 1;
				result = -EIO;
				break;
			} else if (error != 0) {
				TRACEi( 4, "error code   :", error);
				TRACEi( 4, "error command:", command);
			}
		}
		if (status & QIC_STATUS_NEW_CARTRIDGE) {
			int error;
			int command;
#ifndef NO_TRACE_AT_ALL
			int old_tracing = tracing;

			/*  Undocumented feature: Must clear (not present!) error
			 *  here or we'll fail later.
			 */
			tracing = 0;
			ftape_report_error( &error, &command, 1);
			tracing = old_tracing;
#else
			ftape_report_error( &error, &command, 1);
#endif
			TRACE( 4, "status: new cartridge");
			*new_tape = 1;
		}
	} while (status & QIC_STATUS_ERROR);

	*no_tape = !(status & QIC_STATUS_CARTRIDGE_PRESENT);
	*wp_tape = (status & QIC_STATUS_WRITE_PROTECT);
	if (*no_tape) {
		TRACE( 3, "no cartridge present");
	} else {
		if (*wp_tape) {
			TRACE( 3, "Write protected cartridge");
		}
	}
	TRACE_EXIT;
	return result;
}

void
ftape_log_vendor_id( void)
{
	TRACE_FUN( 5, "ftape_log_vendor_id");
	int vendor_index;

	ftape_report_vendor_id( &drive_type.vendor_id);
	vendor_index = lookup_vendor_id( drive_type.vendor_id);
	if (drive_type.vendor_id == UNKNOWN_VENDOR && 
	    drive_type.wake_up == wake_up_colorado) {
		vendor_index = 0;
		drive_type.vendor_id = 0; /* hack to get rid of all this mail */
	}
	if (vendor_index < 0) {
		/* Unknown vendor id, first time opening device.
		 * The drive_type remains set to type found at wakeup time, this
		 * will probably keep the driver operating for this new vendor.
		 */
		TRACE( -1,   "============ unknown vendor id ===========");
		TRACE( -1,   "A new, yet unsupported tape drive is found");
		TRACE( -1,   "Please report the following values:");
		TRACEx1( -1, "   Vendor id     : 0x%04x", drive_type.vendor_id);
		TRACEx1( -1, "   Wakeup method : %s", methods[ drive_type.wake_up].name);
		TRACE( -1,   "And a description of your tape drive to:");
		TRACE(-1,    "Kai Harrekilde-Petersen <khp@dolphinics.no>");
		TRACE( -1,   "==========================================");
		drive_type.speed = 500;     /* deci-ips: very safe value */
	} else {
		drive_type.name = vendors[ vendor_index].name;
		drive_type.speed = vendors[ vendor_index].speed;
		TRACEx1( 4, "tape drive type: %s", drive_type.name);
		/* scan all methods for this vendor_id in table */
		while (drive_type.wake_up != vendors[ vendor_index].wake_up) {
			if (vendor_index < NR_ITEMS( vendors) - 1 &&
			    vendors[ vendor_index + 1].vendor_id == drive_type.vendor_id) {
				++vendor_index;
			} else {
				break;
			}
		}
		if (drive_type.wake_up != vendors[ vendor_index].wake_up) {
			TRACE( -1, "==========================================");
			TRACE( -1, "wakeup type mismatch:");
			TRACEx2( -1, "found: %s, expected: %s",
				 methods[ drive_type.wake_up].name,
				 methods[ vendors[ vendor_index].wake_up].name);
			TRACE(-1, "please report this to <khp@dolphinics.no>");
			TRACE( -1, "==========================================");
		}
	}
	TRACE_EXIT;
}

void
ftape_calc_timeouts( void)
{
	TRACE_FUN( 5, "ftape_calc_timeouts");
	int speed;                    /* deci-ips ! */
	int length;

	/*                           tape transport speed
	 *  data rate:        QIC-40   QIC-80   QIC-3010 QIC-3020
	 *
	 *    250 Kbps        25 ips     n/a      n/a      n/a
	 *    500 Kbps        50 ips   34 ips   22.6 ips   n/a
	 *      1 Mbps          n/a    68 ips   45.2 ips 22.6 ips
	 *      2 Mbps          n/a      n/a      n/a    45.2 ips
	 *
	 *  fast tape transport speed is at least 68 ips.
	 */
	switch (qic_std) {
	case QIC_TAPE_QIC40:
		speed = (ftape_data_rate == 3) ? 250 : 500;
		break;
	case QIC_TAPE_QIC80:
		speed = (ftape_data_rate == 2) ? 340 : 680;
		break;
	case QIC_TAPE_QIC3010:
		speed = (ftape_data_rate == 2) ? 226 : 452;
		break;
	case QIC_TAPE_QIC3020:
		speed = (ftape_data_rate == 1) ? 226 : 452;
		break;
	default:
		TRACE( 1, "Unknown qic_std (bug) ?");
		speed = 500;
		break;
	}
	if (tape_len <= 0) {
		/*  Handle unknown length tapes as 1100 ft ones (worst case)
		 */
		TRACE( 3, "Unknown tape length, using worst case timing values!");
		length = 1100;
	} else {
		length = tape_len;
	}
	if (drive_type.speed == 0) {
		unsigned long t0;
		int dt;

		ftape_seek_to_bot();
		t0 = jiffies;
		ftape_seek_to_eot();
		ftape_seek_to_bot();
		dt = (int) (((jiffies - t0) * USPT) / 1000);
		drive_type.speed = (2 * 12 * length * 1000) / dt;
		TRACE( -1, "==========================================");
		TRACEx1( -1, "drive : %s", drive_type.name);
		TRACEx2( -1, "delta time = %d, length = %d", dt, length);
		TRACEx1( -1, "has max tape speed of %d ips", drive_type.speed);
		TRACE(   -1, "please report this to <khp@dolphinics.no>");
		TRACE(   -1, "==========================================");
	}
	/*  time to go from bot to eot at normal speed (data rate):
	 *  time = (1+delta) * length (ft) * 12 (inch/ft) / speed (ips)
	 *  delta = 10 % for seek speed, 20 % for rewind speed.
	 */
	timeout.seek = (length * 132 * SECOND) / speed;
	timeout.rewind = (length * 144 * SECOND) / (10 * drive_type.speed);
	timeout.reset = 20 * SECOND + timeout.rewind;
	TRACEx2( 4, "speed = %d, length = %d", speed, length);
	TRACEx1( 4, "seek timeout: %d sec", (timeout.seek + 500) / 1000);
	TRACEx1( 4, "rewind timeout: %d sec", (timeout.rewind + 500) / 1000);
	TRACE_EXIT;
}

int
ftape_init_drive( int* formatted)
{
	TRACE_FUN( 5, "ftape_init_drive");
	int result = 0;
	int status;

	result = ftape_report_raw_drive_status( &status);
	if (result >= 0 && (status & QIC_STATUS_CARTRIDGE_PRESENT)) {
		if (!(status & QIC_STATUS_AT_BOT)) {
			/*  Antique drives will get here after a soft reset,
			 *  modern ones only if the driver is loaded when the
			 *  tape wasn't rewound properly.
			 */
			ftape_seek_to_bot();
		}
		if (!(status & QIC_STATUS_REFERENCED)) {
			TRACE( 5, "starting seek_load_point");
			result = ftape_command_wait( QIC_SEEK_LOAD_POINT,
						     timeout.reset, &status);
			if (result < 0) {
				TRACE( 1, "seek_load_point failed (command)");
			}
		}
	}
	if (result >= 0) {
		int rate;

		*formatted = (status & QIC_STATUS_REFERENCED);
		if (!*formatted) {
			TRACE( 2, "tape is not formatted !");
		}
		/*  Select highest rate supported by both fdc and drive.
		 *  Start with highest rate supported by the fdc.
		 */
		if (fdc.type >= i82078_1) {
			rate = 0;
		} else if (fdc.type >= i82077) {
			rate = 1;
		} else {
			rate = 2;
		}
		do {
			result = ftape_set_data_rate( rate);
			if (result >= 0) {
				ftape_calc_timeouts();
				break;
			}
			++rate;
		} while (rate < 4);
		if (result < 0) {
			result = -EIO;
		}
	}
	if (result >= 0) {
		/* Tape should be at bot if new cartridge ! */
		ftape_new_cartridge();
	}
	init_drive_needed = 0;
	TRACE_EXIT;
	return result;
}


static int
def_idle_state( void ) { 
	TRACE_FUN( 5, "def_idle_state");
	int result= 0;
	
	if ( first_data_segment == -1) {
		TRACEi( 5, "first_data_segment: ",first_data_segment);
		ftape_deblock_segment = -1;
		result = read_header_segment( deblock_buffer, 0 );
		TRACEi( 5, "first_data_segment: ",first_data_segment);
		ftape_deblock_segment = -1;
	} else {
		result = ftape_flush_buffers( 0 ); /* "0" says to position before a possibly
						    * created file-mark and thus to stay within
						    * the same volume
						    */
		ftape_zap_write_buffers();
	}
	if ( result >= 0 ) {
		result = ftape_abort_operation();
		if (result < 0) {
			TRACE( 1, "ftape_abort_operation failed");
			result = -EIO;
		}
		/* clear remaining read buffers */
		ftape_zap_read_buffers();
	}
	ftape_state= idle;
	ftape_io_state = io_idle;
	TRACE_EXIT;
	return result;
}

/******************************************************************************
 *
 *  functions for the MTIOCTOP commands
 *
 *****************************************************************************/

static int 
mt_dummy( int *dummy )
{
	TRACE_FUN( 5, "mt_dummy");
	
	TRACE_EXIT;
	return -ENOSYS;
}

static int 
mt_reset( int *dummy )
{        
	TRACE_FUN( 5, "mt_reset");
	int result;
	
	(void)ftape_seek_to_bot();
	result = ftape_reset_drive();
	init_drive_needed = 1;
	ftape_offline = 0;
	ftape_new_cartridge();
	TRACE_EXIT;
	return result ;
}

static int 
mt_fsf( int *arg )
{
	TRACE_FUN( 5, "mt_fsf");
	int result;

	result = qic80_skip_files( qic80_find_file_no(ftape_seg_pos), *arg );
	just_before_eof = 0;
	TRACE_EXIT;
	return result;
}

static int 
mt_bsf( int *arg )
{
	TRACE_FUN( 5, "mt_bsf");
	int result;
	
	result = qic80_skip_files( qic80_find_file_no(ftape_seg_pos), -*arg );
	just_before_eof = 0;
	TRACE_EXIT;
	return result;
}

int
seek_block( int data_offset, int block_increment )
{ 
	TRACE_FUN( 5, "seek_block");
	int result     = 0;
	int act_block  = 0;
	int start_seg, end_seg;
	int vol_size;                                  
	unsigned block_size;
	int new_block_pos = -1;
	int vol_no;
	
	vol_no = qic80_find_file_no ( ftape_seg_pos );
	start_seg = qic80_get_volume_boundaries( vol_no, &end_seg );
	if ( start_seg == 0 || end_seg == 0 ) {
		result = -EIO;
	}
	if ( result >= 0 ) {
		/* user needn't to worry about blocksize if not in raw_mode
		 */
		block_size = qic80_get_volume_block_size( vol_no );
		new_block_pos = data_offset/block_size + block_increment;
		if ( new_block_pos < 0 ) {
			TRACEx1( 3, "new_block_pos %d < 0", new_block_pos);
			new_block_pos = 0;
			result = -EINVAL;
		}
		vol_size = qic80_get_volume_size( vol_no );
		if ( qic80_get_volume_cmpr_flag( vol_no ) ) {
			result = zftape_compress_seek( new_block_pos, 
						       start_seg, end_seg, 
						       block_size );
		} else {
			ftape_uncmpr_pos = block_size * new_block_pos;
			if ( vol_size < ftape_uncmpr_pos ) {
				TRACEx2( 1, "vol_size %d < ftape_uncmpr_pos %d", vol_size, ftape_uncmpr_pos );
				ftape_uncmpr_pos = vol_size;
				result = -EIO;
			}
			ftape_data_pos = ftape_calc_data_pos ( start_seg, 0 )
				+ ftape_uncmpr_pos;
			ftape_seg_pos= 
				ftape_calc_seg_byte_coord( &ftape_seg_data_pos,
							   ftape_data_pos );
		}
		if ( vol_size == ftape_uncmpr_pos ) {
			just_before_eof = 1;
			/*  
			 * why this? because qic80_file_no checks agains 
			 * start and end segment of a volume. We do not want 
			 * to advance to the next volume with this function.
			 */
			TRACE( 4, "set just_before_eof");
			zftape_retard_to_prev_segment();
		} else {
			just_before_eof = 0;
		}
		TRACEi( 4, "new_seg_pos:", ftape_seg_pos);
		TRACEi( 4, "new_data_pos:", ftape_data_pos);
		TRACEi( 4, "vol_size:", vol_size);
		TRACEi( 4, "new_seg_data_pos:", ftape_seg_data_pos);
		act_block = ftape_uncmpr_pos / block_size;
		TRACEi( 4,"block_size: ", block_size );
	}                  
	if ( new_block_pos >= 0 ) {
		ftape_resid = new_block_pos - act_block;
	} else {
		ftape_resid = block_increment;
	}
	if ( ftape_resid < 0 ) {
		ftape_resid = -ftape_resid;
	}
	TRACE_EXIT;
	return result;
}     

static int 
mt_fsr( int *arg )
{ 
	TRACE_FUN( 5, "mt_fsr");
	int result;
	
	result = seek_block( ftape_uncmpr_pos, *arg );
	TRACE_EXIT;
	return result;
}

static int 
mt_bsr( int *arg )
{   
	TRACE_FUN( 5, "mt_bsr");
	int result;
	
	result = seek_block( ftape_uncmpr_pos,  -*arg );
	TRACE_EXIT;
	return result;
}

static int 
mt_weof( int *arg )
{
	TRACE_FUN( 5, "mt_weof");
	int result;
	
	result = qic80_weof( *arg );

	TRACE_EXIT;
	return result;
}

static int 
mt_rew( int *dummy )
{          
	TRACE_FUN( 5, "mt_rew");
	int result = 0;
	
	if( first_data_segment != -1 ) {
		(void)def_idle_state();
	}
	ftape_update_header_segments();
	result = ftape_seek_to_bot();
	ftape_reset_position();
	ftape_zap_read_buffers();
	ftape_zap_write_buffers();
	ftape_resid = 0;
	TRACE_EXIT;
	return result;
}

static int 
mt_offl( int *dummy )
{
	TRACE_FUN( 5, "mt_offl");
	int result;
	
	result = mt_rew( NULL );
	going_offline= 1;
	TRACE_EXIT;
	return result;
}

static int 
mt_nop( int *dummy )
{
	TRACE_FUN( 5, "mt_nop");
	/*  should we set tape status?
	 *
	 */
	/*  if ( !ftape_offline && !no_tape ) (void)def_idle_state( ); */
	TRACE_EXIT;
	return 0; 
}

static int 
mt_reten( int *dummy )
{  
	TRACE_FUN( 5, "mt_reten");
	int result;
	
	if( first_data_segment != -1 ) {
		(void)def_idle_state();
	}
	result = ftape_seek_to_eot();
	if (result >= 0) {
		result= ftape_seek_to_bot();
	}
	ftape_resid = 0;
	TRACE_EXIT;
	return( result );
}

static int 
fsfbsfm( int arg )
{ 
	TRACE_FUN( 5, "fsfbsfm");
	int result = 0;    
	int size;
	int block_pos;
	int block_size;
	int vol_no;
	
	/*
	 * What to do? This should seek to the next file-mark and position 
	 * BEFORE. That is, a next write would just extend the current file. 
	 * Well. Let's just seek to the end of the current file, if count == 1.
	 * If count > 1, then do a "mt_fsf( count - 1 )", and then seek to the
	 * end of that file.
	 *  If count == 0, do nothing
	 */
	if ( arg != 0 ) {
		vol_no = qic80_find_file_no( ftape_seg_pos );
		result = qic80_skip_files( vol_no, arg < 0 ? arg : arg - 1 );
		if ( result >= 0 ) {
			size = qic80_get_volume_size( vol_no );
			block_size = qic80_get_volume_block_size( vol_no );
			block_pos =  (size + block_size - 1) / block_size;
			(void)mt_seek ( &block_pos );
			if ( ftape_uncmpr_pos != size ) {
				/*
				 *   we didn't managed to go there
				 */
				TRACEx2( 1, "wanted position at %d, "
					 "only arrived at %d", 
					 size, ftape_uncmpr_pos );
				ftape_resid = 1;
				result = -EIO;
			} else {
				/*
				 *  allow next write to append data
				 */
				ftape_resid = 0;
				/* just_before_eof = 1;  
				 * already set by seek_block() in case of 
				 * success  
				 */ 
			}
		} else {
			just_before_eof = 0;
			if ( arg > 0 ) {
				ftape_resid++;
			} else {
				ftape_resid--;
			}
		}
	}
	TRACE_EXIT;
	return result;
}   

static int 
mt_bsfm( int *arg )
{
	TRACE_FUN( 5, "mt_bsfm");
	int result;
	
	result = fsfbsfm( -*arg );
	TRACE_EXIT;
	return result;
}

static int 
mt_fsfm( int *arg )
{
	TRACE_FUN( 5, "mt_fsfm");
	int result;
	
	result = fsfbsfm( *arg );
	TRACE_EXIT;
	return result;
}

static int 
mt_eom( int *dummy )
{              
	TRACE_FUN( 5, "mt_eom");
	
	qic80_skip_to_eom();
	ftape_resid = 0;
	TRACE_EXIT;
	return 0;
}

static int 
mt_erase( int *dummy )
{
	TRACE_FUN( 5, "mt_erase");
	int result;
	
	result = zftape_erase();
#ifndef NO_TRACE_AT_ALL
	if ( result < 0 ) {
		TRACE( 1, "zftape_erase failed.");
	}   
#endif
	ftape_resid = 0;
	TRACE_EXIT;
	return result;
}

static int 
mt_ras2( int *dummy ) {
	TRACE_FUN( 5, "mt_ras2");
	int result;
	
	result = -ENOSYS;
	TRACE_EXIT;
	return result;
} 

static int 
mt_setblk( int *new_size )
{
	TRACE_FUN( 5, "mt_setblk");
	int result= 0;
	
	ftape_resid = *new_size;
	if( (*new_size <= 0) || (*new_size > MAX_BLOCK_SIZE ) ) {
#ifndef NO_TRACE_AT_ALL
		TRACEx3( 1, "desired block_size (%d) should "
			 "be >= %d and <= %d bytes.\n",
			 *new_size, 1, MAX_BLOCK_SIZE);
#endif
		result= -EINVAL;
	} else {
		ftape_block_size = *new_size;
	}
	ftape_resid -= ftape_block_size;
	TRACE_EXIT;
	return( result );
} 

static int 
mt_setdensity( int *arg )
{
	TRACE_FUN( 5, "mt_setdensity");
	
#ifndef NO_TRACE_AT_ALL
	tracing = *arg;
	TRACEx1( 4, "tracing set to %d", tracing);
	TRACE_EXIT;
	return 0;
#else
	TRACE_EXIT;
	return -ENOSYS;
#endif
}          

static int 
mt_seek( int *new_block_pos )
{ 
	TRACE_FUN( 8, "mt_seek");
	int result= 0;        
	
	result = seek_block( 0, *new_block_pos );
	TRACE_EXIT;
	return result;
}

/*
 *  return the block position from start of volume
 */
static int
mt_tell( int *arg )
{
	TRACE_FUN( 5, "mt_tell");
	int block_size;
	int vol_no;
	
	vol_no = qic80_find_file_no( ftape_seg_pos );
	block_size = qic80_get_volume_block_size( vol_no );
	*arg = (ftape_uncmpr_pos + block_size - 1) / block_size;
	TRACE_EXIT;
	return 0;
}

/*
 *  change the amount of dma-able memory consumed by the driver.
 *  only usable if compiled with DYN_ALLOC
 *
 */
static int
mt_setdrvbuffer( int *bytes )
{          
	TRACE_FUN( 8, "mt_setdrvbuffer");
	int result= 0;
	
	ftape_resid = *bytes;
	if ( first_data_segment != -1 && !(ftape_offline || no_tape) ) {
		/*
		 * better flush buffers before doing anything to the 
		 * dma-buffers.
		 */
		(void)def_idle_state();
	}
	if ( !suser() ) {
		/* 
		 *  Only root is permitted to change the amount of dma-memory.
		 */
		result = -EACCES;
	} else if ( *bytes % BUFF_SIZE != 0 ) {
		TRACEx2( 1, "new drive buffering (%d) is not a multiple of %d",
			 *bytes,                BUFF_SIZE);
		result= -EINVAL;
	} else {
		result= zftape_change_dma_size( *bytes / BUFF_SIZE );
		/*
		 *  install the new addresses to the buffer struct
		 */
		ftape_zap_read_buffers();
	}
	ftape_resid -= ftape_num_buffers * BUFF_SIZE;
	TRACE_EXIT;
	return result;
}


/*      OPEN routine called by kernel-interface code
 */
int
_ftape_open( void)
{
	TRACE_FUN( 5, "_ftape_open");
	int result;
	static int new_tape = 1;
	
	result = fdc_init();
	if (result >= 0) {
		result = ftape_activate_drive( &drive_type);
		if (result < 0) {
			fdc_disable();
			fdc_release_irq_and_dma();
			
		} else {
			result = ftape_get_drive_status( &new_tape, &no_tape, 
							 &write_protected);
			if (result < 0) {
				ftape_detach_drive();
			} else {
				if (drive_type.vendor_id == UNKNOWN_VENDOR) {
					ftape_log_vendor_id();
				}
				zftape_set_flags( ftape_unit );
				if (no_tape) {
					ftape_offline = 1;
				} else if (new_tape) {
					ftape_offline = 0;
					init_drive_needed = 1;
					read_only = 0;        /* enable writes
							       * again 
							       */
				}
				/*
				 *  now we know if we have a new tape, or no cartridge at all
				 *  and therefor we release the compression map if do not need
				 *  it any longer. It will be reallocated when reading the
				 *  header segments. The compression map occupies up 60K of 
				 *  physical memory for long tapes.
				 *  The disadvantage is that read_header_segments() could
				 *  possibly return with -ENOMEM, but that can hardly occur
				 *  if there is enough swap space. If not, we lose. 
				 */
				if ( new_tape || ftape_offline ) {
					/*
					 *  ftape_offline also includes no_tape
					 */
					zftape_release_compression_map();
					zftape_release_bad_sector_map();
					zftape_release_qic80_vtbl();
				}
				if (!ftape_offline && init_drive_needed) {
					ftape_offline = 0;
					result = ftape_init_drive( &formatted);
					if (result >= 0) {
						new_tape = 0;
					} else {
						ftape_detach_drive();
					}
				}
				if (result >= 0) {
					clear_history();
				}
			}
		}
	}
	TRACEi( 4, "ftape_seg_pos: ", ftape_seg_pos);
	TRACE_EXIT;
	return result;
}

/*      RELEASE routine called by kernel-interface code
 */
int
_ftape_close( void)
{
	TRACE_FUN( 5, "_ftape_close");
	int result = 0;
	int last_segment = 0;
	
	if (!ftape_offline) {
		result = ftape_flush_buffers( 1 ); /* the parameter "1" says 
						    * to position after a 
						    *possibly created file mark
						    */
		last_segment = ftape_seg_pos - 1;
		if ( !(read_only || (ftape_unit & FTAPE_NO_REWIND))) {
			if (result >= 0) {
				result = ftape_update_header_segments();
				if (result < 0) {
					TRACE( 1, "error: update of header "
					       "segments failed");
				}   
			} else {
				TRACE( 1, "error: unable to update "
				       "header segments");
			}
		}
		ftape_abort_operation();
		if (!(ftape_unit & FTAPE_NO_REWIND)) {
			if (!no_tape) {
				TRACE( 5, "rewinding tape");
				result = ftape_seek_to_bot();
			}
			ftape_reset_position();
		}
		ftape_zap_write_buffers();
		/* force call to ftape_zap_read_buffers() and
		 * thus initialization of buffer structure
		 * also, set buf_len_rd to 0
		 */
		ftape_zap_read_buffers();  
		/* high level driver status, forces creation of 
		 * a new volume when calling ftape_write again
		 * and not just_before_eof
		 */
		ftape_io_state = io_idle;  
		ftape_deblock_segment = -1;
	} 
	ftape_detach_drive();
	fdc_uninit();
#ifndef NO_TRACE_AT_ALL
	if (history.used) {
		TRACE( 3,  "== Non-fatal errors this run: ==");
		TRACE( 3,  "fdc isr statistics:");
		TRACEi( 3, " id_am_errors     :", history.id_am_errors);
		TRACEi( 3, " id_crc_errors    :", history.id_crc_errors);
		TRACEi( 3, " data_am_errors   :", history.data_am_errors);
		TRACEi( 3, " data_crc_errors  :", history.data_crc_errors);
		TRACEi( 3, " overrun_errors   :", history.overrun_errors);
		TRACEi( 3, " no_data_errors   :", history.no_data_errors);
		TRACEi( 3, " retries          :", history.retries);
		if (history.used & 1) {
			TRACE( 3,  "ecc statistics:");
			TRACEi( 3, " crc_errors       :",history.crc_errors);
			TRACEi( 3, " crc_failures     :",history.crc_failures);
			TRACEi( 3, " ecc_failures     :",history.ecc_failures);
			TRACEi( 3, " sectors corrected:",history.corrected);
		}
		TRACEx2( 3, "media defects     : %d%s", history.defects,
			 history.defects ? " !!!" : "");
		TRACEi( 3, "repositions       :", history.rewinds);
		TRACEi( 3, "last segment      :", last_segment);
		if ( history.wr_uncompressed != 0 ) {
			TRACE(  3, "");
			TRACE( 3, "compression statistics (writes):");
			TRACEi(3, "before compression:", 
			       history.wr_uncompressed);
			TRACEi( 3, "after  compression:", 
				history.wr_compressed );
			if ( history.wr_compressed > (1<<14) ) {
				TRACEx1( 3,"compr./uncmpr.    : %d %% ",
					 ( (history.wr_compressed>>10) * 100)
					 / (history.wr_uncompressed>>10)   );
			} else {
				TRACEx1( 3,"compr./uncmpr.    : %d %% ",
					 (history.wr_compressed * 100)
					 / history.wr_uncompressed   );
			}
		}
		if ( history.rd_uncompressed != 0 )
			{
				TRACE(  3, "");
				TRACE(  3, "compression statistics (reads)");
				TRACEi( 3, "before decompression:", 
					history.rd_compressed );
				TRACEi( 3, "after  decompression:", 
					history.rd_uncompressed );
				if ( history.rd_compressed > (1<<14) ) {
					TRACEx1( 3, "compr./uncmpr.      : %d %% ",
						 ( (history.rd_compressed>>10) * 100 )
						 / (history.rd_uncompressed>>10) );
				} else {
					TRACEx1( 3,"compr./uncmpr.      : %d %% ",
						 ( history.rd_compressed * 100 )
						 / history.rd_uncompressed );
				}
			}
	}
#endif /* NO_TRACE_AT_ALL */
	if (going_offline || no_tape ) {
		going_offline = 0;
		ftape_offline = 1;
		zftape_release_compression_map();
		zftape_release_bad_sector_map();
		zftape_release_qic80_vtbl();
	}
	TRACE_EXIT;
	return result;
}

static int
mtioctop_ioctl( struct mtop *mtop, int arg_size )
{
	TRACE_FUN( 5, "mtioctop_ioctl");
	int result = 0;
	fun_entry *mt_fun_entry;
	
	TRACEx1( 5, "calling MTIOCTOP command: 0x%08x", mtop->mt_op);
	if ( arg_size != sizeof( struct mtop ) ) {
		TRACEi( 1, "bad argument size", arg_size);
		TRACE_EXIT;
		return -EINVAL;
	}
	if( mtop->mt_op >= NR_MT_CMDS ) {
		result= -EINVAL;
	} else {
		mt_fun_entry= &mt_funs[ mtop->mt_op ];
		if ( ( (ftape_offline || no_tape) && !mt_fun_entry->offline) ||
		     (!formatted && !mt_fun_entry->not_formatted)) {
#ifndef NO_TRACE_AT_ALL
			if ( ftape_offline ){TRACE(1,"drive is offline");     }
			else if ( no_tape ) {TRACE(1,"no tape present");      }
			else                {TRACE(1,"tape is not formatted");}
#endif
			result = -EPERM;
		} else if (write_protected && !mt_fun_entry->write_protected) {
			TRACE( 1, "drive is write-protected");
			result = -EROFS;
		} else {
			if ( mt_fun_entry->need_idle_state && 
			     !(ftape_offline || no_tape || !formatted) ) {
				result= def_idle_state();
			}
			if( result >= 0 ) {
				if ( !zftape_qic80_mode && 
				     !mt_fun_entry->raw_mode) {
					TRACE( 1, "Drive needs to be in QIC-80 compatibility mode for this command");
					result = -EPERM;
				} else {
					result= (mt_fun_entry->function)( &mtop->mt_count );
				}
			}
		}
	}
	TRACE_EXIT;
	return result;
}

static int
mtiocget_ioctl( struct mtget *mtget, int arg_size )
{
	TRACE_FUN( 5, "mtioctop_ioctl");
	int max_data_pos;
	
	if ( arg_size != sizeof( struct mtget ) ) {
		TRACEi( 1, "bad argument size:", arg_size);
		TRACE_EXIT;
		return -EINVAL;
	}
	mtget->mt_type  = drive_type.vendor_id + 0x800000;
	if ( ftape_report_raw_drive_status( (int *)&mtget->mt_dsreg ) < 0 ) { 
		mtget->mt_dsreg = 0; /* drive type dependent status */
	}
	mtget->mt_erreg = ftape_last_error; /* error register */
	mtget->mt_resid = ftape_resid; /* residuum of writes, reads and 
					* MTIOCTOP commands 
					*/
	if ( !(ftape_offline || no_tape) ) {
		mtget->mt_gstat = GMT_ONLINE(~0UL);
		if (write_protected || ftape_old_ftape) { 
			mtget->mt_gstat |= GMT_WR_PROT(~0UL);
		}
		if( first_data_segment != -1 ) {
			max_data_pos = ftape_total_data_amount;
			if ( zftape_use_compression ) {
				max_data_pos -= ftape_block_size + 8 - 1;
			}
			if ( ftape_data_pos >= max_data_pos ) {
				mtget->mt_gstat |= GMT_EOT(~0UL);
			}
			mtget->mt_fileno = qic80_find_file_no( ftape_seg_pos );
			mtget->mt_blkno=  
				ftape_uncmpr_pos 
				/ qic80_get_volume_block_size( mtget->mt_fileno);
			if ( qic80_check_for_eom( ftape_seg_pos ) ) {
				mtget->mt_gstat |= GMT_EOD(~0UL);
			}
			if ( just_before_eof ) {
				mtget->mt_gstat |= GMT_EOF(~0UL);
			}
			if ( mtget->mt_fileno == 0 && ftape_uncmpr_pos == 0 ) {
				mtget->mt_gstat |= GMT_BOT(~0UL);
			}
		} else {
			mtget->mt_fileno = mtget->mt_blkno = -1;
			if ( mtget->mt_dsreg & QIC_STATUS_AT_BOT ) {
				mtget->mt_gstat |= GMT_BOT(~0UL);
			}
		}
	} else {
		if ( no_tape ) {
			mtget->mt_gstat = GMT_DR_OPEN(~0UL);
		} else {
			mtget->mt_gstat = 0UL;
		}
 		mtget->mt_fileno = mtget->mt_blkno  = -1;
	}
	TRACE_EXIT;
	return 0;
}


static int
mtioc_zftape_setget_seg( unsigned int command, struct mtqic80segment * mtseg, int arg_size )
{
	TRACE_FUN( 5, "mtioc_zftape_setget_seg");
	int result;
	
	TRACE( 5, "Mag tape ioctl command: MTIOC_ZFTAPE_GETSEG");
	if ( zftape_qic80_mode ) {
		TRACE_EXIT;
		return -EPERM;
	} 
	if ( arg_size != sizeof( struct mtqic80segment ) ) {
		TRACEi( 1, "bad argument size:", arg_size);
		TRACE_EXIT;
		return -EINVAL;
	}
	if ( mtseg->data != NULL ) {
		result = verify_area( command == MTIOC_ZFTAPE_GETSEG ? 
				      VERIFY_WRITE : VERIFY_READ, 
				      mtseg->data, 29*SECTOR_SIZE );
		if (result) {
			TRACEx1( 1, "Failed to verify pointer %p "
				 "for writing", mtseg->data );
			TRACE_EXIT;
			return result;
		}
	} else {
		TRACE_EXIT;
		return -EINVAL;
	}
	result = def_idle_state();
	if ( result < 0 ) {
		TRACE_EXIT;
		return result;
	}
	if ( mtseg->seg_num > (segments_per_track * tracks_per_tape) ) {
		TRACE( 1, "segment number is too large");
		TRACE_EXIT;
		return -EINVAL;
	}
	if ( command == MTIOC_ZFTAPE_GETSEG ) {
		result = read_segment( mtseg->seg_num, deblock_buffer, 0);
		if ( result > 0 ) {
			memcpy_tofs( mtseg->data, deblock_buffer, result );
		}      
	} else {
		memcpy_fromfs( deblock_buffer, mtseg->data, 29*SECTOR_SIZE );
		result = _write_segment( mtseg->seg_num, deblock_buffer, 1);
		if (result >= 0 && runner_status == idle) {
			/* Force flush for single segment instead of relying 
			 * on flush in read_segment for multiple segments.
			 */
			result = start_writing( WRITE_SINGLE );
			if (result >= 0 && ftape_state == writing) {
				result = loop_until_writes_done();
				prevent_flush();
			}
		}
	}
	mtseg->result = result;
	/* a negativ result is not an ioctl error 
	 * if the user wants to read damaged tapes, it's
	 * up to him
	 */
	TRACE_EXIT;
	return 0;
}
  
static int 
mtioc_zftape_getblksz( struct mtblksz *blksz, int arg_size )
{
	TRACE_FUN( 5, "mtioc_zftape_getblksz");
	int result;
	int vol_no;
	
	TRACE( 5, "Mag tape ioctl command: MTIOC_ZFTAPE_GETBLKSZ");
	if ( arg_size != sizeof( struct mtblksz ) ) {
		TRACEi( 1, "bad argument size:", arg_size);
		TRACE_EXIT;
		return -EINVAL;
	}
	result = def_idle_state();
	if ( result >= 0 ) {
		vol_no = qic80_find_file_no(ftape_seg_pos);
		blksz->mt_blksz = qic80_get_volume_block_size( vol_no );
	}
	TRACE_EXIT;
	return 0;
}

static int
mtiocpos_ioctl( struct mtpos *mtpos, int arg_size )
{
	TRACE_FUN( 5, "mtiocpos_ioctl");
	
	TRACE( 5, "Mag tape ioctl command: MTIOCPOS");
	if ( arg_size != sizeof( struct mtpos ) ) {
		TRACEi( 1, "bad argument size:", arg_size);
		TRACE_EXIT;
		return -EINVAL;
	}
	TRACE_EXIT;
	return mt_tell( (int *)&mtpos->mt_blkno );
}

/*      IOCTL routine called by kernel-interface code
 */
int
_ftape_ioctl( unsigned int command, void * arg)
{
	TRACE_FUN( 5, "_ftape_ioctl");
	int result = -EINVAL;
	union { struct mtop mtop;
		struct mtget mtget;
		struct mtpos mtpos;
		struct mtqic80segment mtqic80segment;
		struct mtblksz mtblksz;
	} krnl_arg;
	int arg_size = _IOC_SIZE(command);
	int dir = _IOC_DIR(command);
	int error;
	
	/* This check will only catch arguments that are too large !
	 */
	if (dir & (_IOC_READ | _IOC_WRITE) && arg_size > sizeof( krnl_arg)) {
		TRACEi( 1, "bad argument size:", arg_size);
		TRACE_EXIT;
		return -EINVAL;
	}
	if (dir & _IOC_WRITE) {
		error = verify_area( VERIFY_READ, arg, arg_size);
		if (error) {
			TRACE_EXIT;
			return error;
		}
		memcpy_fromfs( &krnl_arg.mtop, arg, arg_size);
	}
	TRACEx1( 5, "called with ioctl command: 0x%08x", command);
	switch (command) {
	case MTIOCTOP:
		result = mtioctop_ioctl( &krnl_arg.mtop, arg_size );
		break;
	case MTIOCGET:
		result = mtiocget_ioctl( &krnl_arg.mtget, arg_size );
		break;
	case MTIOCPOS:
		result = mtiocpos_ioctl( &krnl_arg.mtpos, arg_size );
		break;
	case MTIOC_ZFTAPE_GETBLKSZ:
		result = mtioc_zftape_getblksz( &krnl_arg.mtblksz, arg_size );
		result = 0;
		break;
	case MTIOC_ZFTAPE_GETSEG: /* read a segment via ioctl */
	case MTIOC_ZFTAPE_SETSEG: /* write a segment via ioctl */
		result = mtioc_zftape_setget_seg( command, 
						  &krnl_arg.mtqic80segment, 
						  arg_size );
		break;
	default:
		result = -EINVAL;
		break;
	}
	if (dir & _IOC_READ) {
		error = verify_area( VERIFY_WRITE, arg, arg_size);
		if (error) {
			TRACE_EXIT;
			return error;
		}
		memcpy_tofs( arg, &krnl_arg, arg_size);
	}
	TRACE_EXIT;
	return result;
}

void
ftape_init_driver( void)
{
	drive_type.vendor_id = UNKNOWN_VENDOR;
	drive_type.speed = 0;
	drive_type.wake_up = unknown_wake_up;
	drive_type.name = "Unknown";

	timeout.seek = 650 * SECOND;
	timeout.reset = 670 * SECOND;
	timeout.rewind = 650 * SECOND;
	timeout.head_seek = 15 * SECOND;
	timeout.stop = 5 * SECOND;
	timeout.pause = 16 * SECOND;
	
	qic_std = -1;
	tape_len = -1;
	current_command = 0;
	current_cylinder = -1;
	
	segments_per_track = 102;
	segments_per_head = 1020;
	segments_per_cylinder = 4;
	tracks_per_tape = 20;
	ftape_failure = 1;
	ftape_seg_pos = 0;
	first_data_segment = -1;
	ftape_state = idle;
	no_tape = 1;
	formatted = 0;
	ftape_data_rate = 0;
	going_offline = 0;
	read_only = 0;
	
	init_drive_needed = 1;
	header_segment_1 = -1;
	header_segment_2 = -1;
	used_header_segment = -1;
	location.track = -1;
	location.known = 0;
	tape_running = 0;
	might_be_off_track = 1;
	
	ftape_new_cartridge();        /* init some tape related variables */
}

